""" This module provides the infrastructure for loopy preambles that have
a complex requirement structure. This includes:
* preambles triggering the creation of other preambles
* a caching mechanism to avoid duplicated preambles where harmful

TODO rename to generation_cache. I will use it for much more than preambles.
"""

from pytools import Record

# Base class for all cache items.
class UFL2LoopyCacheItem(Record):
    __slots__ = ["content", "returnValue"]

    def __init__(self, content):
        self.content = content
        self.returnValue = content

class NoReturnCacheItem(UFL2LoopyCacheItem):
    def __init__(self, content):
        UFL2LoopyCacheItem.__init__(self, content)
        self.returnValue = None

class LoopyInameCacheItem(UFL2LoopyCacheItem):
    pass

class SymbolCacheItem(UFL2LoopyCacheItem):
    pass

class PreambleCacheItem(NoReturnCacheItem):
    counter = 0
    def __init__(self, content):
        NoReturnCacheItem.__init__(self, content)
        self.content = (PreambleCacheItem.counter, content)
        PreambleCacheItem.counter = PreambleCacheItem.counter + 1

class TemporaryVariableCacheItem(UFL2LoopyCacheItem):
    import numpy
    def __init__(self, content, dtype=numpy.float64):
        UFL2LoopyCacheItem.__init__(self, content)
        from loopy import TemporaryVariable
        self.content = TemporaryVariable(content, dtype)
        self.returnVariable = content

class InstructionCacheItem(NoReturnCacheItem):
    def __init__(self, content):
        NoReturnCacheItem.__init__(self, content)

class CInstructionCacheItem(InstructionCacheItem):
    def __init__(self, content, inames=[], assignees=[]):
        InstructionCacheItem.__init__(self, content)
        from loopy import CInstruction
        self.content = CInstruction(inames, content, assignees=assignees)

class ExpressionInstructionCacheItem(InstructionCacheItem):
    pass

class LoopDomainCacheItem(UFL2LoopyCacheItem):
    def __init__(self, content, upperbound=None):
        if upperbound:
            self.content = "{{ [{0}] : 0<={0}<{1} }}".format(content, upperbound)
        else:
            self.content = "{{ [{0}] : 0<={0}<{0}_n }}".format(content)
        self.returnValue = content

class IncludeCacheItem(NoReturnCacheItem):
    def __init__(self, content):
        NoReturnCacheItem.__init__(self, content)
        self.content = "#include<{}>".format(content)


# have one cache the module level. It is easier than handing around an instance of it.
_cache = {}


def freeze(data):
    """ A function that deterministically generates an
    immutable item from the given data. Used for cache key generation.
    """
    from collections import Hashable, Iterable, MutableMapping

    # Strings are iterable and hashable, but they do not pose any problems.
    if isinstance(data, str):
        return data

    import ufl.classes
    if isinstance(data, ufl.classes.Expr):
        return data

    # Check if the given data is already hashable
    if isinstance(data, Hashable):
        if isinstance(data, Iterable):
            return type(data)(freeze(i) for i in data)
        return data

    # convert standard mutable containers
    if isinstance(data, MutableMapping):
        return tuple((freeze(k), freeze(v)) for k,v in data.iteritems())
    if isinstance(data, Iterable):
        return tuple(freeze(i) for i in data)

    # we don't know how to handle this object, so we give up
    raise TypeError('Cannot freeze non-hashable object {} of type {}'.format(data, type(data)))

class _NoCachingCounter(object):
    counter = 0
    def get(self):
        _NoCachingCounter.counter = _NoCachingCounter.counter + 1
        return _NoCachingCounter.counter


def no_caching(*a):
    return _NoCachingCounter.get()


class _RegisteredFunction(object):
    """ The data structure for a function that accesses UFL2LoopyDataCache """
    def __init__(self, func,
                 cache_key_generator=lambda *a : freeze(a),
                 cache_item_type=UFL2LoopyCacheItem,
                 **kwargs
                ):
        self.func = func
        self.cache_key_generator = cache_key_generator
        self.cache_item_type = cache_item_type
        self.kwargs = kwargs

        assert issubclass(cache_item_type, UFL2LoopyCacheItem)

    def __call__(self, *args):
        # Get the cache key from the given arguments
        cache_key = (self, self.cache_key_generator(*args))
        # check whether we have a cache hit
        if cache_key in _cache:
            # and return the result depending on the cache item type
            return _cache[cache_key].returnValue
        else:
            # evaluate the original function and wrap it in a cache item
            content = self.cache_item_type(self.func(*args), **self.kwargs)
            _cache[cache_key] = content
            return content.returnValue


def _dune_decorator_factory(**factory_kwargs):
    """ A function decorator factory

    Generates a function decorator, that turns a given function into
    a function that stores its result in the UFL2LoopyDataCache.

    The generated decorator may be used with or without keyword arguments.

    The keyword arguments given to this factory override any keywords
    given to the decorator. This allows the definition of specialized
    decorators for different purposes.

    Possible keyword arguments:
    ---------------------------
    cache_key_generator : function
        A function that maps the arguments to the function to an immutable cache key.
        Defaults to generate_cache_tuple.
    cache_item_type : type of UFL2LoopyCacheItem
        The type of to wrap the contents in when storing in the cache.

    Any excess keyword arguments will be forwarded to the CacheItem!
    """
    def _dec(*args, **kwargs):
        # Modify the kwargs according to the factorys kwargs
        for k in factory_kwargs:
            kwargs[k] = factory_kwargs[k]
        if args:
            assert len(args) == 1
            return _RegisteredFunction(args[0], **kwargs)
        else:
            return lambda f: _RegisteredFunction(f, **kwargs)

    return _dec

def cache_preambles():
    return [v.content for v in _cache.values() if isinstance(v, PreambleCacheItem)]

def cache_instructions():
    return [v.content for v in _cache.values() if isinstance(v, InstructionCacheItem)]

def cache_temporaries():
    return {v.returnValue: v.content for v in _cache.values() if isinstance(v, TemporaryVariableCacheItem)}

def cache_loop_domains():
    return [v.content for v in _cache.values() if isinstance(v, LoopDomainCacheItem)]

# Define some decorators that will be useful!
loop_domain = _dune_decorator_factory(cache_item_type=LoopDomainCacheItem)
dune_preamble = _dune_decorator_factory(cache_item_type=PreambleCacheItem)
loopy_iname = _dune_decorator_factory(cache_item_type=LoopyInameCacheItem)
dune_symbol = _dune_decorator_factory(cache_item_type=SymbolCacheItem)
temporary_variable = _dune_decorator_factory(cache_item_type=TemporaryVariableCacheItem)
loopy_c_instruction = _dune_decorator_factory(cache_item_type=CInstructionCacheItem)
loopy_expr_instruction = _dune_decorator_factory(cache_item_type=ExpressionInstructionCacheItem)

@_dune_decorator_factory(cache_item_type=IncludeCacheItem)
def dune_include(incfile):
     return incfile